from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class Unit(StrEnum):
    """
    Specifies the physical unit of sensed data reported by SCADA

    Choices and descriptions:

      * Unknown:
      * Unitless:
      * W:
      * Celcius:
      * Fahrenheit:
      * Gpm:
      * Wh:
      * AmpsRms:
      * VoltsRms:
    """

    Unknown = auto()
    Unitless = auto()
    W = auto()
    Celcius = auto()
    Fahrenheit = auto()
    Gpm = auto()
    Wh = auto()
    AmpsRms = auto()
    VoltsRms = auto()

    @classmethod
    def default(cls) -> "Unit":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



""" List of all the schema types """

from gwatn.types.accepted_bid import AcceptedBid
from gwatn.types.accepted_bid import AcceptedBid_Maker
from gwatn.types.atn_bid import AtnBid
from gwatn.types.atn_bid import AtnBid_Maker
from gwatn.types.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore,
)
from gwatn.types.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore_Maker,
)
from gwatn.types.atn_params_report_heatpumpwithbooststore import (
    AtnParamsReportHeatpumpwithbooststore,
)
from gwatn.types.atn_params_report_heatpumpwithbooststore import (
    AtnParamsReportHeatpumpwithbooststore_Maker,
)
from gwatn.types.base_g_node_gt import BaseGNodeGt
from gwatn.types.base_g_node_gt import BaseGNodeGt_Maker
from gwatn.types.basegnode_scada_create import BasegnodeScadaCreate
from gwatn.types.basegnode_scada_create import BasegnodeScadaCreate_Maker
from gwatn.types.discoverycert_algo_create import DiscoverycertAlgoCreate
from gwatn.types.discoverycert_algo_create import DiscoverycertAlgoCreate_Maker
from gwatn.types.dispatch_contract_confirmed_heatpumpwithbooststore import (
    DispatchContractConfirmedHeatpumpwithbooststore,
)
from gwatn.types.dispatch_contract_confirmed_heatpumpwithbooststore import (
    DispatchContractConfirmedHeatpumpwithbooststore_Maker,
)
from gwatn.types.flo_params_heatpumpwithbooststore import (
    FloParamsHeatpumpwithbooststore,
)
from gwatn.types.flo_params_heatpumpwithbooststore import (
    FloParamsHeatpumpwithbooststore_Maker,
)
from gwatn.types.g_node_gt import GNodeGt
from gwatn.types.g_node_gt import GNodeGt_Maker
from gwatn.types.g_node_instance_gt import GNodeInstanceGt
from gwatn.types.g_node_instance_gt import GNodeInstanceGt_Maker
from gwatn.types.gt_dispatch_boolean import GtDispatchBoolean
from gwatn.types.gt_dispatch_boolean import GtDispatchBoolean_Maker
from gwatn.types.gw_cert_id import GwCertId
from gwatn.types.gw_cert_id import GwCertId_Maker
from gwatn.types.heartbeat_a import HeartbeatA
from gwatn.types.heartbeat_a import HeartbeatA_Maker
from gwatn.types.heartbeat_algo_audit import HeartbeatAlgoAudit
from gwatn.types.heartbeat_algo_audit import HeartbeatAlgoAudit_Maker
from gwatn.types.heartbeat_b import HeartbeatB
from gwatn.types.heartbeat_b import HeartbeatB_Maker
from gwatn.types.initial_tadeed_algo_create import InitialTadeedAlgoCreate
from gwatn.types.initial_tadeed_algo_create import InitialTadeedAlgoCreate_Maker
from gwatn.types.initial_tadeed_algo_optin import InitialTadeedAlgoOptin
from gwatn.types.initial_tadeed_algo_optin import InitialTadeedAlgoOptin_Maker
from gwatn.types.initial_tadeed_algo_transfer import InitialTadeedAlgoTransfer
from gwatn.types.initial_tadeed_algo_transfer import InitialTadeedAlgoTransfer_Maker
from gwatn.types.join_dispatch_contract import JoinDispatchContract
from gwatn.types.join_dispatch_contract import JoinDispatchContract_Maker
from gwatn.types.latest_price import LatestPrice
from gwatn.types.latest_price import LatestPrice_Maker
from gwatn.types.market_slot import MarketSlot
from gwatn.types.market_slot import MarketSlot_Maker
from gwatn.types.market_type_gt import MarketTypeGt
from gwatn.types.market_type_gt import MarketTypeGt_Maker
from gwatn.types.new_tadeed_algo_optin import NewTadeedAlgoOptin
from gwatn.types.new_tadeed_algo_optin import NewTadeedAlgoOptin_Maker
from gwatn.types.new_tadeed_send import NewTadeedSend
from gwatn.types.new_tadeed_send import NewTadeedSend_Maker
from gwatn.types.old_tadeed_algo_return import OldTadeedAlgoReturn
from gwatn.types.old_tadeed_algo_return import OldTadeedAlgoReturn_Maker
from gwatn.types.price_quantity import PriceQuantity
from gwatn.types.price_quantity import PriceQuantity_Maker
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless_Maker
from gwatn.types.ready import Ready
from gwatn.types.ready import Ready_Maker
from gwatn.types.scada_cert_transfer import ScadaCertTransfer
from gwatn.types.scada_cert_transfer import ScadaCertTransfer_Maker
from gwatn.types.sim_scada_driver_report import SimScadaDriverReport
from gwatn.types.sim_scada_driver_report import SimScadaDriverReport_Maker
from gwatn.types.sim_timestep import SimTimestep
from gwatn.types.sim_timestep import SimTimestep_Maker
from gwatn.types.sla_enter import SlaEnter
from gwatn.types.sla_enter import SlaEnter_Maker
from gwatn.types.snapshot_heatpumpwithbooststore import SnapshotHeatpumpwithbooststore
from gwatn.types.snapshot_heatpumpwithbooststore import (
    SnapshotHeatpumpwithbooststore_Maker,
)
from gwatn.types.super_starter import SuperStarter
from gwatn.types.super_starter import SuperStarter_Maker
from gwatn.types.supervisor_container_gt import SupervisorContainerGt
from gwatn.types.supervisor_container_gt import SupervisorContainerGt_Maker
from gwatn.types.tadeed_specs_hack import TadeedSpecsHack
from gwatn.types.tadeed_specs_hack import TadeedSpecsHack_Maker
from gwatn.types.tavalidatorcert_algo_create import TavalidatorcertAlgoCreate
from gwatn.types.tavalidatorcert_algo_create import TavalidatorcertAlgoCreate_Maker
from gwatn.types.tavalidatorcert_algo_transfer import TavalidatorcertAlgoTransfer
from gwatn.types.tavalidatorcert_algo_transfer import TavalidatorcertAlgoTransfer_Maker
from gwatn.types.terminalasset_certify_hack import TerminalassetCertifyHack
from gwatn.types.terminalasset_certify_hack import TerminalassetCertifyHack_Maker


__all__ = [
    "PriceQuantity",
    "PriceQuantity_Maker",
    "SimScadaDriverReport",
    "SimScadaDriverReport_Maker",
    "HeartbeatAlgoAudit",
    "HeartbeatAlgoAudit_Maker",
    "TerminalassetCertifyHack",
    "TerminalassetCertifyHack_Maker",
    "TavalidatorcertAlgoCreate",
    "TavalidatorcertAlgoCreate_Maker",
    "NewTadeedAlgoOptin",
    "NewTadeedAlgoOptin_Maker",
    "HeartbeatA",
    "HeartbeatA_Maker",
    "Ready",
    "Ready_Maker",
    "InitialTadeedAlgoOptin",
    "InitialTadeedAlgoOptin_Maker",
    "InitialTadeedAlgoTransfer",
    "InitialTadeedAlgoTransfer_Maker",
    "PriceQuantityUnitless",
    "PriceQuantityUnitless_Maker",
    "TadeedSpecsHack",
    "TadeedSpecsHack_Maker",
    "AtnParamsHeatpumpwithbooststore",
    "AtnParamsHeatpumpwithbooststore_Maker",
    "GNodeInstanceGt",
    "GNodeInstanceGt_Maker",
    "BaseGNodeGt",
    "BaseGNodeGt_Maker",
    "FloParamsHeatpumpwithbooststore",
    "FloParamsHeatpumpwithbooststore_Maker",
    "JoinDispatchContract",
    "JoinDispatchContract_Maker",
    "AcceptedBid",
    "AcceptedBid_Maker",
    "MarketSlot",
    "MarketSlot_Maker",
    "SnapshotHeatpumpwithbooststore",
    "SnapshotHeatpumpwithbooststore_Maker",
    "AtnBid",
    "AtnBid_Maker",
    "SuperStarter",
    "SuperStarter_Maker",
    "InitialTadeedAlgoCreate",
    "InitialTadeedAlgoCreate_Maker",
    "GwCertId",
    "GwCertId_Maker",
    "GtDispatchBoolean",
    "GtDispatchBoolean_Maker",
    "SlaEnter",
    "SlaEnter_Maker",
    "BasegnodeScadaCreate",
    "BasegnodeScadaCreate_Maker",
    "LatestPrice",
    "LatestPrice_Maker",
    "OldTadeedAlgoReturn",
    "OldTadeedAlgoReturn_Maker",
    "SupervisorContainerGt",
    "SupervisorContainerGt_Maker",
    "AtnParamsReportHeatpumpwithbooststore",
    "AtnParamsReportHeatpumpwithbooststore_Maker",
    "HeartbeatB",
    "HeartbeatB_Maker",
    "ScadaCertTransfer",
    "ScadaCertTransfer_Maker",
    "NewTadeedSend",
    "NewTadeedSend_Maker",
    "GNodeGt",
    "GNodeGt_Maker",
    "TavalidatorcertAlgoTransfer",
    "TavalidatorcertAlgoTransfer_Maker",
    "DiscoverycertAlgoCreate",
    "DiscoverycertAlgoCreate_Maker",
    "MarketTypeGt",
    "MarketTypeGt_Maker",
    "SimTimestep",
    "SimTimestep_Maker",
    "DispatchContractConfirmedHeatpumpwithbooststore",
    "DispatchContractConfirmedHeatpumpwithbooststore_Maker",
]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class GNodeRole(StrEnum):
    """
    Categorizes GNodes by their function within GridWorks. [More Info](https://gridworks.readthedocs.io/en/latest/g-node-role.html).

    Choices and descriptions:

      * GNode: Default value
      * TerminalAsset: An avatar for a real-word Transactive Device. [More Info](https://gridworks.readthedocs.io/en/latest/transactive-device.html).
      * AtomicTNode: Transacts in markets on behalf of, and controlling the power use of, a TerminalAsset
      * MarketMaker: Runs energy markets at its Node in the GNodeTree
      * AtomicMeteringNode: Role of a GNode that will become an AtomicTNode, prior to it owning TaTradingRights
      * ConductorTopologyNode: An avatar for a real-world electric grid node - e.g. a substation or transformer
      * InterconnectionComponent: An avatar for a cable or wire on the electric grid
      * World: Adminstrative GNode responsible for managing and authorizing instances
      * TimeCoordinator: Responsible for managing time in simulations
      * Supervisor: Responsible for GNode actors running in a container
      * Scada: GNode associated to the device and code that directly monitors and actuates a Transactive Device
      * PriceService: Provides price forecasts for markets run by MarketMakers
      * WeatherService: Provides weather forecasts
      * AggregatedTNode: An aggregation of AtomicTNodes
    """

    GNode = auto()
    TerminalAsset = auto()
    Scada = auto()
    PriceService = auto()
    WeatherService = auto()
    AggregatedTNode = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    World = auto()
    TimeCoordinator = auto()
    Supervisor = auto()

    @classmethod
    def default(cls) -> "GNodeRole":
        """
        Returns default value GNode
        """
        return cls.GNode

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Dev Utils for Millinocket Demo"""
from gwatn.dev_utils.dev_discoverer import DevDiscoverer
from gwatn.dev_utils.dev_ta_owner import DevTaOwner
from gwatn.dev_utils.dev_validator import DevValidator


__all__ = [
    "DevDiscoverer",
    "DevTaOwner",
    "DevValidator",
]



"""Type supervisor.container.gt, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import SupervisorContainerStatus


class SupervisorContainerStatus000SchemaEnum:
    enum_name: str = "supervisor.container.status.000"
    symbols: List[str] = [
        "00000000",
        "f48cff43",
        "17c5cc54",
        "ec342324",
        "cfde1b40",
        "4e28b6ae",
        "da2dafe0",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class SupervisorContainerStatus000(StrEnum):
    Unknown = auto()
    Authorized = auto()
    Launching = auto()
    Provisioning = auto()
    Running = auto()
    Stopped = auto()
    Deleted = auto()

    @classmethod
    def default(cls) -> "SupervisorContainerStatus000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class SupervisorContainerStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> SupervisorContainerStatus:
        if not SupervisorContainerStatus000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to SupervisorContainerStatus000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum,
            SupervisorContainerStatus,
            SupervisorContainerStatus.default(),
        )

    @classmethod
    def local_to_type(
        cls, supervisor_container_status: SupervisorContainerStatus
    ) -> str:
        if not isinstance(supervisor_container_status, SupervisorContainerStatus):
            raise SchemaError(
                f"{supervisor_container_status} must be of type {SupervisorContainerStatus}"
            )
        versioned_enum = as_enum(
            supervisor_container_status,
            SupervisorContainerStatus000,
            SupervisorContainerStatus000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, SupervisorContainerStatus000] = {
        "00000000": SupervisorContainerStatus000.Unknown,
        "f48cff43": SupervisorContainerStatus000.Authorized,
        "17c5cc54": SupervisorContainerStatus000.Launching,
        "ec342324": SupervisorContainerStatus000.Provisioning,
        "cfde1b40": SupervisorContainerStatus000.Running,
        "4e28b6ae": SupervisorContainerStatus000.Stopped,
        "da2dafe0": SupervisorContainerStatus000.Deleted,
    }

    versioned_enum_to_type_dict: Dict[SupervisorContainerStatus000, str] = {
        SupervisorContainerStatus000.Unknown: "00000000",
        SupervisorContainerStatus000.Authorized: "f48cff43",
        SupervisorContainerStatus000.Launching: "17c5cc54",
        SupervisorContainerStatus000.Provisioning: "ec342324",
        SupervisorContainerStatus000.Running: "cfde1b40",
        SupervisorContainerStatus000.Stopped: "4e28b6ae",
        SupervisorContainerStatus000.Deleted: "da2dafe0",
    }


def check_is_world_instance_name_format(v: str) -> None:
    """
    WorldInstanceName format: A single alphanumerical word starting
    with an alphabet char (the root GNodeAlias) and an integer,
    seperated by '__'. For example 'd1__1'

    Raises:
        ValueError: if not WorldInstanceNameFormat format
    """
    try:
        words = v.split("__")
    except:
        raise ValueError(f"{v} is not split by '__'")
    if len(words) != 2:
        raise ValueError(f"{v} not 2 words separated by '__'")
    try:
        int(words[1])
    except:
        raise ValueError(f"{v} second word not an int")

    root_g_node_alias = words[0]
    first_char = root_g_node_alias[0]
    if not first_char.isalpha():
        raise ValueError(f"{v} first word must be alph char")
    if not root_g_node_alias.isalnum():
        raise ValueError(f"{v} first word must be alphanumeric")


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class SupervisorContainerGt(BaseModel):
    """Used to send and receive updates about SupervisorContainers.

    Sent from a GNodeRegistry to a World, and used also by the World as it spawns
    GNodeInstances in docker instances (i.e., the SupervisorContainers).
    [More info](https://gridworks.readthedocs.io/en/latest/supervisor.html).
    """

    SupervisorContainerId: str = Field(
        title="Id of the docker SupervisorContainer",
    )
    Status: SupervisorContainerStatus = Field(
        title="Status",
    )
    WorldInstanceName: str = Field(
        title="Name of the WorldInstance",
        description="For example, d1__1 is a potential name for a World whose World GNode has alias d1.",
    )
    SupervisorGNodeInstanceId: str = Field(
        title="Id of the SupervisorContainer's prime actor (aka the Supervisor GNode)",
    )
    SupervisorGNodeAlias: str = Field(
        title="Alias of the SupervisorContainer's prime actor (aka the Supervisor GNode)",
    )
    TypeName: Literal["supervisor.container.gt"] = "supervisor.container.gt"
    Version: str = "000"

    @validator("SupervisorContainerId")
    def _check_supervisor_container_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"SupervisorContainerId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("Status")
    def _check_status(cls, v: SupervisorContainerStatus) -> SupervisorContainerStatus:
        return as_enum(v, SupervisorContainerStatus, SupervisorContainerStatus.Unknown)

    @validator("WorldInstanceName")
    def _check_world_instance_name(cls, v: str) -> str:
        try:
            check_is_world_instance_name_format(v)
        except ValueError as e:
            raise ValueError(
                f"WorldInstanceName failed WorldInstanceNameFormat format validation: {e}"
            )
        return v

    @validator("SupervisorGNodeInstanceId")
    def _check_supervisor_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"SupervisorGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("SupervisorGNodeAlias")
    def _check_supervisor_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"SupervisorGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Status"]
        Status = as_enum(
            self.Status, SupervisorContainerStatus, SupervisorContainerStatus.default()
        )
        d["StatusGtEnumSymbol"] = SupervisorContainerStatusMap.local_to_type(Status)
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class SupervisorContainerGt_Maker:
    type_name = "supervisor.container.gt"
    version = "000"

    def __init__(
        self,
        supervisor_container_id: str,
        status: SupervisorContainerStatus,
        world_instance_name: str,
        supervisor_g_node_instance_id: str,
        supervisor_g_node_alias: str,
    ):
        self.tuple = SupervisorContainerGt(
            SupervisorContainerId=supervisor_container_id,
            Status=status,
            WorldInstanceName=world_instance_name,
            SupervisorGNodeInstanceId=supervisor_g_node_instance_id,
            SupervisorGNodeAlias=supervisor_g_node_alias,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: SupervisorContainerGt) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> SupervisorContainerGt:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SupervisorContainerGt:
        d2 = dict(d)
        if "SupervisorContainerId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainerId")
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in SupervisorContainerStatus000SchemaEnum.symbols:
            d2["Status"] = SupervisorContainerStatusMap.type_to_local(
                d2["StatusGtEnumSymbol"]
            )
        else:
            d2["Status"] = SupervisorContainerStatus.default()
        if "WorldInstanceName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing WorldInstanceName")
        if "SupervisorGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorGNodeInstanceId")
        if "SupervisorGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorGNodeAlias")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return SupervisorContainerGt(
            SupervisorContainerId=d2["SupervisorContainerId"],
            Status=d2["Status"],
            WorldInstanceName=d2["WorldInstanceName"],
            SupervisorGNodeInstanceId=d2["SupervisorGNodeInstanceId"],
            SupervisorGNodeAlias=d2["SupervisorGNodeAlias"],
            TypeName=d2["TypeName"],
            Version="000",
        )



""" List of all the types used"""
from typing import Dict
from typing import List
from typing import no_type_check

from gwatn.types import AcceptedBid_Maker
from gwatn.types import AtnBid_Maker
from gwatn.types import AtnParamsHeatpumpwithbooststore_Maker
from gwatn.types import AtnParamsReportHeatpumpwithbooststore_Maker
from gwatn.types import BaseGNodeGt_Maker
from gwatn.types import BasegnodeScadaCreate_Maker
from gwatn.types import DiscoverycertAlgoCreate_Maker
from gwatn.types import DispatchContractConfirmedHeatpumpwithbooststore_Maker
from gwatn.types import FloParamsHeatpumpwithbooststore_Maker
from gwatn.types import GNodeGt_Maker
from gwatn.types import GNodeInstanceGt_Maker
from gwatn.types import GtDispatchBoolean_Maker
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatAlgoAudit_Maker
from gwatn.types import HeartbeatB_Maker
from gwatn.types import InitialTadeedAlgoCreate_Maker
from gwatn.types import InitialTadeedAlgoOptin_Maker
from gwatn.types import InitialTadeedAlgoTransfer_Maker
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import LatestPrice_Maker
from gwatn.types import MarketSlot_Maker
from gwatn.types import MarketTypeGt_Maker
from gwatn.types import NewTadeedAlgoOptin_Maker
from gwatn.types import NewTadeedSend_Maker
from gwatn.types import OldTadeedAlgoReturn_Maker
from gwatn.types import PriceQuantity_Maker
from gwatn.types import PriceQuantityUnitless_Maker
from gwatn.types import Ready_Maker
from gwatn.types import ScadaCertTransfer_Maker
from gwatn.types import SimScadaDriverReport_Maker
from gwatn.types import SimTimestep_Maker
from gwatn.types import SlaEnter_Maker
from gwatn.types import SnapshotHeatpumpwithbooststore_Maker
from gwatn.types import SuperStarter_Maker
from gwatn.types import SupervisorContainerGt_Maker
from gwatn.types import TadeedSpecsHack_Maker
from gwatn.types import TavalidatorcertAlgoCreate_Maker
from gwatn.types import TavalidatorcertAlgoTransfer_Maker
from gwatn.types import TerminalassetCertifyHack_Maker


TypeMakerByName: Dict[str, HeartbeatA_Maker] = {}


@no_type_check
def type_makers() -> List[HeartbeatA_Maker]:
    return [
        AcceptedBid_Maker,
        AtnBid_Maker,
        AtnParamsHeatpumpwithbooststore_Maker,
        AtnParamsReportHeatpumpwithbooststore_Maker,
        BaseGNodeGt_Maker,
        BasegnodeScadaCreate_Maker,
        DiscoverycertAlgoCreate_Maker,
        DispatchContractConfirmedHeatpumpwithbooststore_Maker,
        FloParamsHeatpumpwithbooststore_Maker,
        GNodeGt_Maker,
        GNodeInstanceGt_Maker,
        GtDispatchBoolean_Maker,
        GwCertId_Maker,
        HeartbeatA_Maker,
        HeartbeatAlgoAudit_Maker,
        HeartbeatB_Maker,
        InitialTadeedAlgoCreate_Maker,
        InitialTadeedAlgoOptin_Maker,
        InitialTadeedAlgoTransfer_Maker,
        JoinDispatchContract_Maker,
        LatestPrice_Maker,
        MarketSlot_Maker,
        MarketTypeGt_Maker,
        NewTadeedAlgoOptin_Maker,
        NewTadeedSend_Maker,
        OldTadeedAlgoReturn_Maker,
        PriceQuantity_Maker,
        PriceQuantityUnitless_Maker,
        Ready_Maker,
        ScadaCertTransfer_Maker,
        SimScadaDriverReport_Maker,
        SimTimestep_Maker,
        SlaEnter_Maker,
        SnapshotHeatpumpwithbooststore_Maker,
        SuperStarter_Maker,
        SupervisorContainerGt_Maker,
        TadeedSpecsHack_Maker,
        TavalidatorcertAlgoCreate_Maker,
        TavalidatorcertAlgoTransfer_Maker,
        TerminalassetCertifyHack_Maker,
    ]


for maker in type_makers():
    TypeMakerByName[maker.type_name] = maker


def version_by_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are TypeNames, values are versions
    """

    v: Dict[str, str] = {
        "accepted.bid": "000",
        "atn.bid": "001",
        "atn.params.heatpumpwithbooststore": "000",
        "atn.params.report.heatpumpwithbooststore": "000",
        "base.g.node.gt": "002",
        "basegnode.scada.create": "000",
        "discoverycert.algo.create": "000",
        "dispatch.contract.confirmed.heatpumpwithbooststore": "000",
        "flo.params.heatpumpwithbooststore": "000",
        "g.node.gt": "002",
        "g.node.instance.gt": "000",
        "gt.dispatch.boolean": "110",
        "gw.cert.id": "000",
        "heartbeat.a": "100",
        "heartbeat.algo.audit": "000",
        "heartbeat.b": "000",
        "initial.tadeed.algo.create": "000",
        "initial.tadeed.algo.optin": "002",
        "initial.tadeed.algo.transfer": "000",
        "join.dispatch.contract": "000",
        "latest.price": "000",
        "market.slot": "000",
        "market.type.gt": "000",
        "new.tadeed.algo.optin": "000",
        "new.tadeed.send": "000",
        "old.tadeed.algo.return": "000",
        "price.quantity": "000",
        "price.quantity.unitless": "000",
        "ready": "001",
        "scada.cert.transfer": "000",
        "sim.scada.driver.report": "000",
        "sim.timestep": "000",
        "sla.enter": "000",
        "snapshot.heatpumpwithbooststore": "000",
        "super.starter": "000",
        "supervisor.container.gt": "000",
        "tadeed.specs.hack": "000",
        "tavalidatorcert.algo.create": "000",
        "tavalidatorcert.algo.transfer": "000",
        "terminalasset.certify.hack": "000",
    }

    return v


def status_by_versioned_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are versioned TypeNames, values are type status
    """

    v: Dict[str, str] = {
        "accepted.bid.000": "Pending",
        "atn.bid.001": "Pending",
        "atn.params.heatpumpwithbooststore.000": "Active",
        "atn.params.report.heatpumpwithbooststore.000": "Pending",
        "base.g.node.gt.002": "Pending",
        "basegnode.scada.create.000": "Pending",
        "discoverycert.algo.create.000": "Pending",
        "dispatch.contract.confirmed.heatpumpwithbooststore.000": "Pending",
        "flo.params.heatpumpwithbooststore.000": "Active",
        "g.node.gt.002": "Pending",
        "g.node.instance.gt.000": "Pending",
        "gt.dispatch.boolean.110": "Pending",
        "gw.cert.id.000": "Active",
        "heartbeat.a.100": "Pending",
        "heartbeat.algo.audit.000": "Pending",
        "heartbeat.b.000": "Pending",
        "initial.tadeed.algo.create.000": "Active",
        "initial.tadeed.algo.optin.002": "Active",
        "initial.tadeed.algo.transfer.000": "Active",
        "join.dispatch.contract.000": "Active",
        "latest.price.000": "Pending",
        "market.slot.000": "Pending",
        "market.type.gt.000": "Pending",
        "new.tadeed.algo.optin.000": "Active",
        "new.tadeed.send.000": "Active",
        "old.tadeed.algo.return.000": "Active",
        "price.quantity.000": "Pending",
        "price.quantity.unitless.000": "Pending",
        "ready.001": "Pending",
        "scada.cert.transfer.000": "Pending",
        "sim.scada.driver.report.000": "Pending",
        "sim.timestep.000": "Pending",
        "sla.enter.000": "Pending",
        "snapshot.heatpumpwithbooststore.000": "Pending",
        "super.starter.000": "Pending",
        "supervisor.container.gt.000": "Pending",
        "tadeed.specs.hack.000": "Pending",
        "tavalidatorcert.algo.create.000": "Active",
        "tavalidatorcert.algo.transfer.000": "Active",
        "terminalasset.certify.hack.000": "Pending",
    }

    return v



"""Type heartbeat.a, version 100"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


def check_is_hex_char(v: str) -> None:
    """
    HexChar format: single-char string in '0123456789abcdefABCDEF'

    Raises:
        ValueError: if not HexChar format
    """
    if not isinstance(v, str):
        raise ValueError(f"{v} must be a hex char, but not even a string")
    if len(v) > 1:
        raise ValueError(f"{v} must be a hex char, but not of len 1")
    if v not in "0123456789abcdefABCDEF":
        raise ValueError(f"{v} must be one of '0123456789abcdefABCDEF'")


class HeartbeatA(BaseModel):
    """Used to check that an actor can both send and receive messages.

    Payload for direct messages sent back and forth between actors,
    for example a Supervisor and one of its subordinates.

    [More info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html).
    """

    MyHex: str = Field(
        title="Hex character getting sent",
        default="0",
    )
    YourLastHex: str = Field(
        title="Last hex character received from heartbeat partner",
        default="0",
    )
    TypeName: Literal["heartbeat.a"] = "heartbeat.a"
    Version: str = "100"

    @validator("MyHex")
    def _check_my_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"MyHex failed HexChar format validation: {e}")
        return v

    @validator("YourLastHex")
    def _check_your_last_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"YourLastHex failed HexChar format validation: {e}")
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class HeartbeatA_Maker:
    type_name = "heartbeat.a"
    version = "100"

    def __init__(self, my_hex: str, your_last_hex: str):
        self.tuple = HeartbeatA(
            MyHex=my_hex,
            YourLastHex=your_last_hex,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> HeartbeatA:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA:
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MyHex")
        if "YourLastHex" not in d2.keys():
            raise SchemaError(f"dict {d2} missing YourLastHex")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return HeartbeatA(
            MyHex=d2["MyHex"],
            YourLastHex=d2["YourLastHex"],
            TypeName=d2["TypeName"],
            Version="100",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MarketTypeName(StrEnum):
    """
    Categorizes different markets run by MarketMaker

    Choices and descriptions:

      * unknown: Default unknown
      * rt5gate5: Real-time energy, 5 minute MarketSlots, gate closing 5 minutes prior to start
      * rt60gate5: Real-time energy, 60 minute MarketSlots, gate closing 5 minutes prior to start
      * da60: Day-ahead energy, 60 minute MarketSlots
      * rt60gate30: Real-time energy, 60 minute MarketSlots, gate closing 30 minutes prior to start
      * rt15gate5: Real-time energy, 15 minute MarketSlots, gate closing 5 minutes prior to start
      * rt30gate5: Real-time energy, 30 minute MarketSlots, gate closing 5 minutes prior to start
      * rt60gate30b: Real-time energy, 30 minute MarketSlots, gate closing 5 minutes prior to start, QuantityUnit AvgkW
    """

    unknown = auto()
    rt5gate5 = auto()
    rt60gate5 = auto()
    da60 = auto()
    rt60gate30 = auto()
    rt15gate5 = auto()
    rt30gate5 = auto()
    rt60gate30b = auto()

    @classmethod
    def default(cls) -> "MarketTypeName":
        """
        Returns default value unknown
        """
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



""" DijkstraSolarFirebricksAlpha Class Definition """
from abc import ABC
from typing import Optional


class DEdge(ABC):
    def __init__(
        self,
        start_ts_idx: int,
        start_idx: int,
        end_idx: int,
        cost: Optional[float] = None,
    ):
        self.start_ts_idx = start_ts_idx
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.cost = cost

    def __repr__(self) -> str:
        return f"DEdge => StartTimeSliceIdx: {self.start_ts_idx}. StartStoreIdx: {self.start_idx}. EndStoreIdx: {self.end_idx}. Cost: {self.cost}"



"""Type g.node.gt, version 002"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.data_classes.g_node import GNode
from gwatn.enums import GNodeRole
from gwatn.enums import GNodeStatus


class GNodeRole000SchemaEnum:
    enum_name: str = "g.node.role.000"
    symbols: List[str] = [
        "00000000",
        "bdeaa0b1",
        "8021dcad",
        "304890c5",
        "8eb5b9e1",
        "234cfaa2",
        "fec0c127",
        "3901c7d2",
        "c499943c",
        "88112a93",
        "674ad859",
        "2161739f",
        "1dce1efd",
        "db57d184",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class GNodeRole000(StrEnum):
    GNode = auto()
    TerminalAsset = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    World = auto()
    TimeCoordinator = auto()
    Supervisor = auto()
    Scada = auto()
    PriceService = auto()
    WeatherService = auto()
    AggregatedTNode = auto()

    @classmethod
    def default(cls) -> "GNodeRole000":
        return cls.GNode

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class GNodeRoleMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> GNodeRole:
        if not GNodeRole000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to GNodeRole000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, GNodeRole, GNodeRole.default())

    @classmethod
    def local_to_type(cls, g_node_role: GNodeRole) -> str:
        if not isinstance(g_node_role, GNodeRole):
            raise SchemaError(f"{g_node_role} must be of type {GNodeRole}")
        versioned_enum = as_enum(g_node_role, GNodeRole000, GNodeRole000.default())
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, GNodeRole000] = {
        "00000000": GNodeRole000.GNode,
        "bdeaa0b1": GNodeRole000.TerminalAsset,
        "8021dcad": GNodeRole000.AtomicTNode,
        "304890c5": GNodeRole000.MarketMaker,
        "8eb5b9e1": GNodeRole000.AtomicMeteringNode,
        "234cfaa2": GNodeRole000.ConductorTopologyNode,
        "fec0c127": GNodeRole000.InterconnectionComponent,
        "3901c7d2": GNodeRole000.World,
        "c499943c": GNodeRole000.TimeCoordinator,
        "88112a93": GNodeRole000.Supervisor,
        "674ad859": GNodeRole000.Scada,
        "2161739f": GNodeRole000.PriceService,
        "1dce1efd": GNodeRole000.WeatherService,
        "db57d184": GNodeRole000.AggregatedTNode,
    }

    versioned_enum_to_type_dict: Dict[GNodeRole000, str] = {
        GNodeRole000.GNode: "00000000",
        GNodeRole000.TerminalAsset: "bdeaa0b1",
        GNodeRole000.AtomicTNode: "8021dcad",
        GNodeRole000.MarketMaker: "304890c5",
        GNodeRole000.AtomicMeteringNode: "8eb5b9e1",
        GNodeRole000.ConductorTopologyNode: "234cfaa2",
        GNodeRole000.InterconnectionComponent: "fec0c127",
        GNodeRole000.World: "3901c7d2",
        GNodeRole000.TimeCoordinator: "c499943c",
        GNodeRole000.Supervisor: "88112a93",
        GNodeRole000.Scada: "674ad859",
        GNodeRole000.PriceService: "2161739f",
        GNodeRole000.WeatherService: "1dce1efd",
        GNodeRole000.AggregatedTNode: "db57d184",
    }


class GNodeStatus100SchemaEnum:
    enum_name: str = "g.node.status.100"
    symbols: List[str] = [
        "00000000",
        "153d3475",
        "a2cfc2f7",
        "839b38db",
        "f5831e1d",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class GNodeStatus100(StrEnum):
    Unknown = auto()
    Pending = auto()
    Active = auto()
    PermanentlyDeactivated = auto()
    Suspended = auto()

    @classmethod
    def default(cls) -> "GNodeStatus100":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class GNodeStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> GNodeStatus:
        if not GNodeStatus100SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to GNodeStatus100 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, GNodeStatus, GNodeStatus.default())

    @classmethod
    def local_to_type(cls, g_node_status: GNodeStatus) -> str:
        if not isinstance(g_node_status, GNodeStatus):
            raise SchemaError(f"{g_node_status} must be of type {GNodeStatus}")
        versioned_enum = as_enum(
            g_node_status, GNodeStatus100, GNodeStatus100.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, GNodeStatus100] = {
        "00000000": GNodeStatus100.Unknown,
        "153d3475": GNodeStatus100.Pending,
        "a2cfc2f7": GNodeStatus100.Active,
        "839b38db": GNodeStatus100.PermanentlyDeactivated,
        "f5831e1d": GNodeStatus100.Suspended,
    }

    versioned_enum_to_type_dict: Dict[GNodeStatus100, str] = {
        GNodeStatus100.Unknown: "00000000",
        GNodeStatus100.Pending: "153d3475",
        GNodeStatus100.Active: "a2cfc2f7",
        GNodeStatus100.PermanentlyDeactivated: "839b38db",
        GNodeStatus100.Suspended: "f5831e1d",
    }


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


class GNodeGt(BaseModel):
    """Used to send and receive updates about GNodes.

    GNodes are the building blocks of Gridworks. They have slowly-changing state
    that must be kept in sync across a distributed system. Therefore, they require
    a global registry to act as Single Source of Truth (SSoT). This class is used for
    that SSoT to share information with actors about their GNodes, and the GNodes
    that they will observe and communicate with.
    [More info](https://gridworks.readthedocs.io/en/latest/g-node.html).
    """

    GNodeId: str = Field(
        title="Immutable identifier for GNode",
    )
    Alias: str = Field(
        title="Structured mutable identifier for GNode",
        description="The GNode Aliases are used for organizing how actors in Gridworks communicate. Together, they also encode the known topology of the electric grid. [More info](https://gridworks.readthedocs.io/en/latest/g-node-alias.html).",
    )
    Status: GNodeStatus = Field(
        title="Lifecycle indicator",
    )
    Role: GNodeRole = Field(
        title="Role within Gridworks",
        description=" [More info](https://gridworks.readthedocs.io/en/latest/g-node-role.html).",
    )
    GNodeRegistryAddr: str = Field(
        title="Algorand address for GNodeRegistry",
        description="For actors in a Gridworks world, the GNodeRegistry is the Single Source of Truth for existence and updates to GNodes. [More info](https://gridworks.readthedocs.io/en/latest/g-node-registry.html).",
    )
    PrevAlias: Optional[str] = Field(
        title="Previous GNodeAlias",
        description="As the topology of the grid updates, GNodeAliases will change to reflect that. This may happen a handful of times over the life of a GNode.",
        default=None,
    )
    GpsPointId: Optional[str] = Field(
        title="Lat/lon of GNode",
        description="Some GNodes, in particular those acting as avatars for physical devices that are part of or are attached to the electric grid, have physical locations. These locations are used to help validate the grid topology.",
        default=None,
    )
    OwnershipDeedId: Optional[int] = Field(
        title="Algorand Id of ASA Deed",
        description="The Id of the TaDeed Algorand Standard Asset if the GNode is a TerminalAsset. [More info](https://gridworks.readthedocs.io/en/latest/ta-deed.html).",
        default=None,
    )
    OwnershipDeedValidatorAddr: Optional[str] = Field(
        title="Algorand address of Validator",
        description="Deeds are issued by the GNodeFactory, in partnership with third party Validators. [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).",
        default=None,
    )
    OwnerAddr: Optional[str] = Field(
        title="Algorand address of the deed owner",
        default=None,
    )
    DaemonAddr: Optional[str] = Field(
        title="Algorand address of the daemon app",
        description="Some GNodes have Daemon applications associated to them to handle blockchain operations.",
        default=None,
    )
    TradingRightsId: Optional[int] = Field(
        title="Algorand Id of ASA TradingRights",
        description="The Id of the TradingRights Algorand Standard Asset.",
        default=None,
    )
    ScadaAlgoAddr: Optional[str] = Field(
        title="ScadaAlgoAddr",
        default=None,
    )
    ScadaCertId: Optional[int] = Field(
        title="ScadaCertId",
        default=None,
    )
    ComponentId: Optional[str] = Field(
        title="Unique identifier for GNode's Component",
        description="Used if a GNode is an avatar for a physical device. The serial number of a device is different from its make/model. The ComponentId captures the specific instance of the device.",
        default=None,
    )
    DisplayName: Optional[str] = Field(
        title="Display Name",
        default=None,
    )
    TypeName: Literal["g.node.gt"] = "g.node.gt"
    Version: str = "002"

    @validator("GNodeId")
    def _check_g_node_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("Alias")
    def _check_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"Alias failed LeftRightDot format validation: {e}")
        return v

    @validator("Status")
    def _check_status(cls, v: GNodeStatus) -> GNodeStatus:
        return as_enum(v, GNodeStatus, GNodeStatus.Unknown)

    @validator("Role")
    def _check_role(cls, v: GNodeRole) -> GNodeRole:
        return as_enum(v, GNodeRole, GNodeRole.GNode)

    @validator("GNodeRegistryAddr")
    def _check_g_node_registry_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeRegistryAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("PrevAlias")
    def _check_prev_alias(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"PrevAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("GpsPointId")
    def _check_gps_point_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GpsPointId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("OwnershipDeedValidatorAddr")
    def _check_ownership_deed_validator_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnershipDeedValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("OwnerAddr")
    def _check_owner_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnerAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("DaemonAddr")
    def _check_daemon_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"DaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("ScadaAlgoAddr")
    def _check_scada_algo_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ScadaAlgoAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("ComponentId")
    def _check_component_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"ComponentId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Status"]
        Status = as_enum(self.Status, GNodeStatus, GNodeStatus.default())
        d["StatusGtEnumSymbol"] = GNodeStatusMap.local_to_type(Status)
        del d["Role"]
        Role = as_enum(self.Role, GNodeRole, GNodeRole.default())
        d["RoleGtEnumSymbol"] = GNodeRoleMap.local_to_type(Role)
        if d["PrevAlias"] is None:
            del d["PrevAlias"]
        if d["GpsPointId"] is None:
            del d["GpsPointId"]
        if d["OwnershipDeedId"] is None:
            del d["OwnershipDeedId"]
        if d["OwnershipDeedValidatorAddr"] is None:
            del d["OwnershipDeedValidatorAddr"]
        if d["OwnerAddr"] is None:
            del d["OwnerAddr"]
        if d["DaemonAddr"] is None:
            del d["DaemonAddr"]
        if d["TradingRightsId"] is None:
            del d["TradingRightsId"]
        if d["ScadaAlgoAddr"] is None:
            del d["ScadaAlgoAddr"]
        if d["ScadaCertId"] is None:
            del d["ScadaCertId"]
        if d["ComponentId"] is None:
            del d["ComponentId"]
        if d["DisplayName"] is None:
            del d["DisplayName"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class GNodeGt_Maker:
    type_name = "g.node.gt"
    version = "002"

    def __init__(
        self,
        g_node_id: str,
        alias: str,
        status: GNodeStatus,
        role: GNodeRole,
        g_node_registry_addr: str,
        prev_alias: Optional[str],
        gps_point_id: Optional[str],
        ownership_deed_id: Optional[int],
        ownership_deed_validator_addr: Optional[str],
        owner_addr: Optional[str],
        daemon_addr: Optional[str],
        trading_rights_id: Optional[int],
        scada_algo_addr: Optional[str],
        scada_cert_id: Optional[int],
        component_id: Optional[str],
        display_name: Optional[str],
    ):
        self.tuple = GNodeGt(
            GNodeId=g_node_id,
            Alias=alias,
            Status=status,
            Role=role,
            GNodeRegistryAddr=g_node_registry_addr,
            PrevAlias=prev_alias,
            GpsPointId=gps_point_id,
            OwnershipDeedId=ownership_deed_id,
            OwnershipDeedValidatorAddr=ownership_deed_validator_addr,
            OwnerAddr=owner_addr,
            DaemonAddr=daemon_addr,
            TradingRightsId=trading_rights_id,
            ScadaAlgoAddr=scada_algo_addr,
            ScadaCertId=scada_cert_id,
            ComponentId=component_id,
            DisplayName=display_name,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: GNodeGt) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> GNodeGt:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GNodeGt:
        d2 = dict(d)
        if "GNodeId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeId")
        if "Alias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Alias")
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in GNodeStatus100SchemaEnum.symbols:
            d2["Status"] = GNodeStatusMap.type_to_local(d2["StatusGtEnumSymbol"])
        else:
            d2["Status"] = GNodeStatus.default()
        if "RoleGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoleGtEnumSymbol")
        if d2["RoleGtEnumSymbol"] in GNodeRole000SchemaEnum.symbols:
            d2["Role"] = GNodeRoleMap.type_to_local(d2["RoleGtEnumSymbol"])
        else:
            d2["Role"] = GNodeRole.default()
        if "GNodeRegistryAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeRegistryAddr")
        if "PrevAlias" not in d2.keys():
            d2["PrevAlias"] = None
        if "GpsPointId" not in d2.keys():
            d2["GpsPointId"] = None
        if "OwnershipDeedId" not in d2.keys():
            d2["OwnershipDeedId"] = None
        if "OwnershipDeedValidatorAddr" not in d2.keys():
            d2["OwnershipDeedValidatorAddr"] = None
        if "OwnerAddr" not in d2.keys():
            d2["OwnerAddr"] = None
        if "DaemonAddr" not in d2.keys():
            d2["DaemonAddr"] = None
        if "TradingRightsId" not in d2.keys():
            d2["TradingRightsId"] = None
        if "ScadaAlgoAddr" not in d2.keys():
            d2["ScadaAlgoAddr"] = None
        if "ScadaCertId" not in d2.keys():
            d2["ScadaCertId"] = None
        if "ComponentId" not in d2.keys():
            d2["ComponentId"] = None
        if "DisplayName" not in d2.keys():
            d2["DisplayName"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return GNodeGt(
            GNodeId=d2["GNodeId"],
            Alias=d2["Alias"],
            Status=d2["Status"],
            Role=d2["Role"],
            GNodeRegistryAddr=d2["GNodeRegistryAddr"],
            PrevAlias=d2["PrevAlias"],
            GpsPointId=d2["GpsPointId"],
            OwnershipDeedId=d2["OwnershipDeedId"],
            OwnershipDeedValidatorAddr=d2["OwnershipDeedValidatorAddr"],
            OwnerAddr=d2["OwnerAddr"],
            DaemonAddr=d2["DaemonAddr"],
            TradingRightsId=d2["TradingRightsId"],
            ScadaAlgoAddr=d2["ScadaAlgoAddr"],
            ScadaCertId=d2["ScadaCertId"],
            ComponentId=d2["ComponentId"],
            DisplayName=d2["DisplayName"],
            TypeName=d2["TypeName"],
            Version="002",
        )

    @classmethod
    def tuple_to_dc(cls, t: GNodeGt) -> GNode:
        if t.GNodeId in GNode.by_id.keys():
            dc = GNode.by_id[t.GNodeId]
        else:
            dc = GNode(
                g_node_id=t.GNodeId,
                alias=t.Alias,
                status=t.Status,
                role=t.Role,
                g_node_registry_addr=t.GNodeRegistryAddr,
                prev_alias=t.PrevAlias,
                gps_point_id=t.GpsPointId,
                ownership_deed_id=t.OwnershipDeedId,
                ownership_deed_validator_addr=t.OwnershipDeedValidatorAddr,
                owner_addr=t.OwnerAddr,
                daemon_addr=t.DaemonAddr,
                trading_rights_id=t.TradingRightsId,
                scada_algo_addr=t.ScadaAlgoAddr,
                scada_cert_id=t.ScadaCertId,
                component_id=t.ComponentId,
                display_name=t.DisplayName,
            )

        return dc

    @classmethod
    def dc_to_tuple(cls, dc: GNode) -> GNodeGt:
        t = GNodeGt_Maker(
            g_node_id=dc.g_node_id,
            alias=dc.alias,
            status=dc.status,
            role=dc.role,
            g_node_registry_addr=dc.g_node_registry_addr,
            prev_alias=dc.prev_alias,
            gps_point_id=dc.gps_point_id,
            ownership_deed_id=dc.ownership_deed_id,
            ownership_deed_validator_addr=dc.ownership_deed_validator_addr,
            owner_addr=dc.owner_addr,
            daemon_addr=dc.daemon_addr,
            trading_rights_id=dc.trading_rights_id,
            scada_algo_addr=dc.scada_algo_addr,
            scada_cert_id=dc.scada_cert_id,
            component_id=dc.component_id,
            display_name=dc.display_name,
        ).tuple
        return t

    @classmethod
    def type_to_dc(cls, t: str) -> GNode:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: GNode) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> GNode:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))



import json
from typing import List
from typing import Optional

import boto3
import pendulum
from gwproto.messages import GtDispatchBoolean
from gwproto.messages import GtDispatchBoolean_Maker
from gwproto.messages import GtShStatus
from gwproto.messages import GtShStatus_Maker
from gwproto.messages import GtTelemetry
from gwproto.messages import GtTelemetry_Maker
from gwproto.messages import SnapshotSpaceheat
from gwproto.messages import SnapshotSpaceheat_Maker
from pydantic import BaseModel

from gwatn.api_types_hack import HackTypeMakerByName


OUT_STUB = "output_data/scada_report_a"


class FileNameMeta(BaseModel):
    FromGNodeAlias: str
    PayloadTypeName: str
    UnixTimeMs: int
    FileName: str
    TypeName: str = "file.name.meta.000"


class StatusOutputRow(BaseModel):
    TimeUnixMs: int
    IntTimeUnixS: int
    Milliseconds: int
    TimeUtc: str
    Value: Optional[int]
    TelemetryName: Optional[str]
    AboutShNode: str
    FromShNode: Optional[str]
    DispatchCmd: Optional[int]
    TypeName: str = "ui.status.output.row.000"


class SnapshotWithSendTime(BaseModel):
    SendTimeUnixMs: int
    Snapshot: SnapshotSpaceheat
    TypeName: str = "wrapped.snapshot.spaceheat.000"


class ScadaReportA_Maker:
    ATN_ALIAS_LIST = ["hw1.isone.ct.newhaven.orange1"]

    def __init__(self, out_stub=OUT_STUB):
        self.s3 = boto3.client("s3")
        self.aws_bucket_name = "gwdev"
        self.world_instance_name = "hw1__1"
        self.out_stub = f"{out_stub}"
        self.output_type_name = "atn.ui.000"
        self.processed_file_name_meta_list = List[FileNameMeta]
        self.new_file_name_meta_list = List[FileNameMeta]
        self.latest_status_list: List[GtShStatus] = []
        self.latest_snapshot_list: List[SnapshotSpaceheat] = []
        self.latest_dispatch_list: List[GtDispatchBoolean] = []

        self.status_rows_to_write: List[StatusOutputRow] = []
        print(f"Initialized {self.__class__}")

    def on_gw_message(self, from_g_node_alias: str, payload):
        if from_g_node_alias not in self.ATN_ALIAS_LIST:
            print(f"Does not record {from_g_node_alias}.")
        else:
            if isinstance(payload, GtShStatus):
                self.latest_status_list.append(payload)

    def get_atn_alias_from_alias(self, alias: str):
        x = alias.split(".")
        if x[-1] == "scada":
            return ".".join(x[:-2])
        else:
            return alias

    def get_status_rows_from_telemetry(
        self, from_sh_node_alias: str, payload: GtTelemetry
    ) -> List[StatusOutputRow]:
        time_unix_ms = payload.ScadaReadTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        row = StatusOutputRow(
            TimeUnixMs=time_unix_ms,
            IntTimeUnixS=int(time_unix_ms / 1000),
            Milliseconds=int(time_unix_ms) % 1000,
            TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
            Value=payload.Value,
            TelemetryName=payload.Name.value,
            AboutShNode=from_sh_node_alias,
            FromShNode=from_sh_node_alias,
        )
        return [row]

    def get_status_rows_from_dispatch(
        self, payload: GtDispatchBoolean
    ) -> List[StatusOutputRow]:
        rows = []
        time_unix_ms = payload.SendTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        row = StatusOutputRow(
            TimeUnixMs=time_unix_ms,
            IntTimeUnixS=int(time_unix_ms / 1000),
            Milliseconds=int(time_unix_ms) % 1000,
            TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
            AboutShNode=payload.AboutNodeAlias,
            FromShNode=payload.FromGNodeAlias,
            DispatchCmd=payload.RelayState,
        )
        rows.append(row)
        return rows

    def get_status_rows_from_snapshot(
        self, payload: SnapshotWithSendTime
    ) -> List[StatusOutputRow]:
        rows = []
        from_g_node_alias = payload.Snapshot.FromGNodeAlias
        snapshot = payload.Snapshot.Snapshot
        time_unix_ms = payload.SendTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        for i in range(len(snapshot.ValueList)):
            row = StatusOutputRow(
                TimeUnixMs=time_unix_ms,
                IntTimeUnixS=int(time_unix_ms / 1000),
                Milliseconds=int(time_unix_ms) % 1000,
                TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                Value=snapshot.ValueList[i],
                TelemetryName=snapshot.TelemetryNameList[i].value,
                AboutShNode=snapshot.AboutNodeAliasList[i],
                FromShNode=from_g_node_alias,
            )
            rows.append(row)
        return rows

    def get_status_rows_from_status(self, payload: GtShStatus) -> List[StatusOutputRow]:
        rows: List[StatusOutputRow] = []
        for single in payload.SimpleTelemetryList:
            for i in range(len(single.ValueList)):
                time_unix_ms = single.ReadTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    Value=single.ValueList[i],
                    TelemetryName=single.TelemetryName.value,
                    AboutShNode=single.ShNodeAlias,
                    FromShNode=single.ShNodeAlias,
                )
                rows.append(row)

        for cmd in payload.BooleanactuatorCmdList:
            about_node_alias = cmd.ShNodeAlias
            for i in range(len(cmd.RelayStateCommandList)):
                time_unix_ms = cmd.CommandTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    AboutShNode=about_node_alias,
                    DispatchCmd=cmd.RelayStateCommandList[i],
                )
                rows.append(row)

        for multi in payload.MultipurposeTelemetryList:
            for i in range(len(multi.ValueList)):
                time_unix_ms = multi.ReadTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    Value=multi.ValueList[i],
                    TelemetryName=multi.TelemetryName.value,
                    AboutShNode=multi.AboutNodeAlias,
                    FromShNode=multi.SensorNodeAlias,
                )
                rows.append(row)
        return rows

    def get_payload_from_s3(self, file_name_meta: FileNameMeta) -> GtShStatus:
        s3_object = self.s3.get_object(
            Bucket=self.aws_bucket_name, Key=file_name_meta.FileName
        )
        gw_type = s3_object["Body"].read()
        try:
            payload_as_dict = json.loads(gw_type)["Payload"]
        except:
            payload_as_dict = json.loads(gw_type)
        maker = HackTypeMakerByName[file_name_meta.PayloadTypeName]
        payload = maker.dict_to_tuple(payload_as_dict)
        if maker.type_alias == SnapshotSpaceheat_Maker.type_alias:
            payload = SnapshotWithSendTime(
                SendTimeUnixMs=file_name_meta.UnixTimeMs, Snapshot=payload
            )
        return payload

    def has_this_days_folder(self, time_s: int) -> bool:
        d = pendulum.from_timestamp(time_s)
        this_days_folder_name = d.strftime("%Y%m%d")
        prefix = f"{self.world_instance_name}/eventstore/{this_days_folder_name}"

        r = self.s3.list_objects_v2(Bucket=self.aws_bucket_name, Prefix=prefix)
        if "Contents" in r.keys():
            return True
        return False

    def get_date_folder_list(
        self, start_time_unix_ms: int, duration_hrs: int
    ) -> List[str]:
        start_s = start_time_unix_ms / 1000
        folder_list: List[str] = []
        found_latest_earlier: bool = False
        i = 0
        while (not found_latest_earlier) and i < 5:
            t = start_s - 3600 * 24 * (i + 1)
            if self.has_this_days_folder(t):
                folder_list.append(pendulum.from_timestamp(t).strftime("%Y%m%d"))
                found_latest_earlier = True
            i += 1

        if self.has_this_days_folder(start_s):
            folder_list.append(pendulum.from_timestamp(start_s).strftime("%Y%m%d"))

        add_hrs = 0
        while add_hrs < duration_hrs:
            add_hrs += 24
            add_hrs = min(add_hrs, duration_hrs)
            t = start_s + add_hrs * 3600
            if self.has_this_days_folder(t):
                folder_list.append(pendulum.from_timestamp(t).strftime("%Y%m%d"))

        return list(set(folder_list))

    def get_file_name_meta_list(
        self,
        start_time_unix_ms: int,
        end_time_unix_ms: int,
        date_folder_list: List[str],
        g_node_alias_list: List[str],
        type_name_list: List[str] = list(HackTypeMakerByName.keys()),
    ):
        fn_list: List[FileNameMeta] = []
        for date_folder in date_folder_list:
            prefix = f"{self.world_instance_name}/eventstore/{date_folder}/"
            paginator = self.s3.get_paginator("list_objects_v2")
            pages = paginator.paginate(Bucket=self.aws_bucket_name, Prefix=prefix)
            file_name_list = []
            for page in pages:
                for obj in page["Contents"]:
                    file_name_list.append(obj["Key"])

            for file_name in file_name_list:
                try:
                    from_g_node_alias = file_name.split("/")[-1].split("-")[0]
                    payload_type_name = file_name.split("/")[-1].split("-")[1]
                    payload_unix_time_ms = int(file_name.split("/")[-1].split("-")[2])
                except:
                    raise Exception(f"Failed file name parsing with {file_name}")

                if (
                    from_g_node_alias in g_node_alias_list
                    and payload_type_name in type_name_list
                    and payload_unix_time_ms > start_time_unix_ms - 300_000
                    and payload_unix_time_ms < end_time_unix_ms
                ):
                    fn_list.append(
                        FileNameMeta(
                            FromGNodeAlias=from_g_node_alias,
                            PayloadTypeName=payload_type_name,
                            UnixTimeMs=payload_unix_time_ms,
                            FileName=file_name,
                        )
                    )

        return fn_list

    def get_csv_rows(
        self, start_time_unix_ms: int, duration_hrs: int, atn_alias: str
    ) -> List[StatusOutputRow]:
        g_node_alias_list = [atn_alias, atn_alias + ".ta.scada"]
        hack_list = [
            "a.garage.temp1",
            "a.tank.out.temp1",
            "a.tank.in.temp1",
            "a.tank.temp0",
        ]
        g_node_alias_list += hack_list
        start_time_utc = pendulum.from_timestamp(start_time_unix_ms / 1000)
        end_time_utc = start_time_utc + pendulum.duration(hours=duration_hrs)
        end_time_unix_ms = end_time_utc.int_timestamp * 1000
        date_folder_list = self.get_date_folder_list(start_time_unix_ms, duration_hrs)
        fn_list = self.get_file_name_meta_list(
            start_time_unix_ms=start_time_unix_ms,
            end_time_unix_ms=end_time_unix_ms,
            date_folder_list=date_folder_list,
            g_node_alias_list=g_node_alias_list,
        )

        rows: List[StatusOutputRow] = []
        for i in range(len(fn_list)):
            fn = fn_list[i]
            if fn.PayloadTypeName == GtDispatchBoolean_Maker.type_alias:
                payload = self.get_payload_from_s3(fn)
                rows += self.get_status_rows_from_dispatch(payload)
            elif fn.PayloadTypeName == GtShStatus_Maker.type_alias:
                payload = self.get_payload_from_s3(fn)
                rows += self.get_status_rows_from_status(payload)
            elif fn.PayloadTypeName == SnapshotSpaceheat_Maker.type_alias:
                payload = self.get_payload_from_s3(fn)
                rows += self.get_status_rows_from_snapshot(payload)
            elif fn.PayloadTypeName == GtTelemetry_Maker.type_alias:
                payload = self.get_payload_from_s3(fn)
                rows += self.get_status_rows_from_telemetry(
                    from_sh_node_alias=fn.FromGNodeAlias,
                    payload=payload,
                )

        rows = sorted(rows, key=lambda x: x.TimeUnixMs)
        return rows

    def make_csv(
        self,
        start_s,
        duration_hrs: int = 48,
        atn_alias: str = "hw1.isone.ct.newhaven.orange1",
    ):
        s = start_s - (start_s % 3600)
        start_time_utc = pendulum.from_timestamp(s)
        print(f"starting at {start_time_utc.strftime('%Y-%m-%d %H:%M')}")
        start_time_unix_ms = s * 1000

        rows = self.get_csv_rows(
            start_time_unix_ms=start_time_unix_ms,
            duration_hrs=duration_hrs,
            atn_alias=atn_alias,
        )
        lines = [
            "TimeUtc, TimeUnixS, Ms, Value, TelemetryName, AboutShNode, FromShNode, DispatchCmd\n"
        ]
        for row in rows:
            line = f"{row.TimeUtc}, {row.IntTimeUnixS}, {row.Milliseconds}"
            if row.Value is None:
                line += ", "
            else:
                line += f",{row.Value}"
            if row.TelemetryName is None:
                line += ", "
            else:
                line += f", {row.TelemetryName}"
            line += f", {row.AboutShNode}"
            if row.FromShNode is None:
                line += ", "
            else:
                line += f", {row.FromShNode}"
            if row.DispatchCmd is None:
                line += ", \n"
            else:
                line += f", {row.DispatchCmd} \n"
            lines.append(line)

        atn_end = atn_alias.split(".")[-1]
        file_name = f"{self.out_stub}/{atn_end}-{start_time_utc.strftime('%Y%m%d-%H%M')}-{duration_hrs}-{atn_alias}.csv"
        print(f"Writing {file_name}")
        with open(file_name, "w") as outfile:
            outfile.writelines(lines)



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MessageCategory(StrEnum):
    """
    Categorizes how GridWorks messages are sent and decoded/encoded

    Choices and descriptions:

      * Unknown: Default value
      * RabbitJsonDirect: Serialized Json message sent on the world rabbit broker from one GNode actor to another
      * RabbitJsonBroadcast: Serailized Json message broadcast on the world rabbit broker by a GNode actor
      * RabbitGwSerial: GwSerial protocol message sent on the world rabbit broker
      * MqttJsonBroadcast: Serialized Json message following MQTT topic format, sent on the world rabbit broker
      * RestApiPost: REST API post
      * RestApiPostResponse: REST API post response
      * RestApiGet: REST API GET
    """

    Unknown = auto()
    RabbitJsonDirect = auto()
    RabbitJsonBroadcast = auto()
    RabbitGwSerial = auto()
    MqttJsonBroadcast = auto()
    RestApiPost = auto()
    RestApiPostResponse = auto()
    RestApiGet = auto()

    @classmethod
    def default(cls) -> "MessageCategory":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type tavalidatorcert.algo.create, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
import gridworks.gw_config as config
from algosdk import encoding
from algosdk.future import transaction
from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class TavalidatorcertAlgoCreate(BaseModel):
    """Used for Step 1 of TaValidator certification.

    Meant to be sent from a pending TaValidator  to the GNodeFactory (Gnf), to
    initiate the process of certifying the pending TaValidator.
    [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).
    """

    ValidatorAddr: str = Field(
        title="The address of the pending TaValidator",
    )
    HalfSignedCertCreationMtx: str = Field(
        title="Algo multi-transaction for certificate creation, with 1 of 2 signatures",
    )
    TypeName: Literal["tavalidatorcert.algo.create"] = "tavalidatorcert.algo.create"
    Version: str = "000"

    @validator("ValidatorAddr")
    def check_validator_addr(cls, v: str) -> str:
        """
        Axiom 5: Uniqueness.
        There must not already be a TaValidatorCert belonging to the  2-sig [GnfAdminAddr, ValidatorAddr] address.
        """
        ValidatorAddr = v
        gnf_admin_addr = config.Public().gnf_admin_addr
        multi = algo_utils.MultisigAccount(
            version=1,
            threshold=2,
            addresses=[gnf_admin_addr, ValidatorAddr],
        )
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        if api_utils.is_validator_addr(ValidatorAddr):
            ValueError("Axiom 5: Already has TaValidatorCert")
        return v

    @validator("HalfSignedCertCreationMtx")
    def check_half_signed_cert_creation_mtx(cls, v: str) -> str:
        """
        Axioms 2, 3:

        Axiom 2: Is AssetConfigTxn.
        The transaction must have type AssetConfigTxn.

        Axiom 3: Is ValidatorCert.
        For the asset getting created: Total is 1, Decimals is 0, UnitName is VLDTR, Manager is GnfAdminAddr,
        AssetName is not blank.
        [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate)
        """
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"HalfSignedCertCreationMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        mtx = encoding.future_msgpack_decode(v)
        txn = mtx.transaction
        gnf_admin_addr = config.Public().gnf_admin_addr
        if not isinstance(txn, transaction.AssetConfigTxn):
            raise ValueError(
                "Axiom 2: The transaction must have type AssetConfigTxn"
                f" got {type(txn)}."
            )
        if not txn.total == 1:
            raise ValueError(
                "Axiom 3: TaValidatorCert ASA Total must be 1, "
                f" got {txn.total}."
                " See https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate"
            )
        if not txn.decimals == 0:
            raise ValueError(
                "Axiom 3: TaValidatorCert ASA Decimals must be 0, "
                f" got {txn.decimals}."
                " See https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate"
            )
        if not txn.unit_name == "VLDTR":
            raise ValueError(
                "Axiom 3: TaValidatorCert ASA UnitName must be 'VLDTR', "
                f" got {txn.decimals}."
                " See https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate"
            )
        if not txn.manager == gnf_admin_addr:
            raise ValueError(
                "Axiom 3: TaValidatorCert manager must be Universe gnf_admin_addr, "
                f" got {txn.decimals}."
                " See https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate"
            )
        if (txn.asset_name is None) or (txn.asset_name == ""):
            raise ValueError(
                "Axiom 3: TaValidatorCert AssetName cannot be blank, "
                " See https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded HalfSignedCertCreationMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr], signed by ValidatorAddr.
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        mtx = encoding.future_msgpack_decode(v.get("HalfSignedCertCreationMtx", None))
        ValidatorAddr = v.get("ValidatorAddr")
        gnf_admin_addr = config.Public().gnf_admin_addr
        multi = algo_utils.MultisigAccount(
            version=1,
            threshold=2,
            addresses=[gnf_admin_addr, ValidatorAddr],
        )

        if not isinstance(mtx, transaction.MultisigTransaction):
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertCreationMtx must have type MultisigTransaction,"
                f" got {type(mtx)}."
            )

        if not multi.addr == mtx.multisig.address():
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertCreationMtx must come from the 2-sig MultiAccount ,"
                f" [GnfAdminAddr, ValidatorAddr], got {type(mtx)}."
            )
        if mtx.multisig.subsigs[1].signature is None:
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertCreationMtx missing TaValidator signature.}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class TavalidatorcertAlgoCreate_Maker:
    type_name = "tavalidatorcert.algo.create"
    version = "000"

    def __init__(self, validator_addr: str, half_signed_cert_creation_mtx: str):
        self.tuple = TavalidatorcertAlgoCreate(
            ValidatorAddr=validator_addr,
            HalfSignedCertCreationMtx=half_signed_cert_creation_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: TavalidatorcertAlgoCreate) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> TavalidatorcertAlgoCreate:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> TavalidatorcertAlgoCreate:
        d2 = dict(d)
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "HalfSignedCertCreationMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HalfSignedCertCreationMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return TavalidatorcertAlgoCreate(
            ValidatorAddr=d2["ValidatorAddr"],
            HalfSignedCertCreationMtx=d2["HalfSignedCertCreationMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



import csv
import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import gridworks.property_format as property_format
import numpy as np
from gridworks.errors import SchemaError


class CsvWeatherForecast(NamedTuple):
    LocationAlias: str
    Comment: str
    StartYearUtc: int
    StartMonthUtc: int
    StartDayUtc: int
    StartHourUtc: int
    UniformSliceDurationHrs: float
    TimezoneString: str
    WeatherUid: str
    TempUnit: Optional[str]
    ForecastUnixTimeS: int
    IrradianceType: Optional[str] = None
    TempSourceAlias: Optional[str] = None
    TempMethodAlias: Optional[str] = None
    IrradianceSourceAlias: Optional[str] = None
    IrradianceMethodAlias: Optional[str] = None
    SkyClaritySourceAlias: Optional[str] = None
    SkyClarityMethodAlias: Optional[str] = None
    Temperatures: Optional[list] = None
    IrradianceWPerM2: Optional[List[int]] = None
    OpaqueCloudCoveragePercents: Optional[List[int]] = None
    TypeName: str = "csv.weather.forecast.sync"

    def asdict(self):
        d = self._asdict()
        if d["IrradianceType"] is None:
            del d["IrradianceType"]
        if d["IrradianceSourceAlias"] is None:
            del d["IrradianceSourceAlias"]
        if d["SkyClarityMethodAlias"] is None:
            del d["SkyClarityMethodAlias"]
        if d["TempSourceAlias"] is None:
            del d["TempSourceAlias"]
        if d["TempMethodAlias"] is None:
            del d["TempMethodAlias"]
        if d["IrradianceWPerM2"] is None:
            del d["IrradianceWPerM2"]
        if d["TempUnit"] is None:
            del d["TempUnit"]
        if d["SkyClaritySourceAlias"] is None:
            del d["SkyClaritySourceAlias"]
        if d["IrradianceMethodAlias"] is None:
            del d["IrradianceMethodAlias"]
        if d["OpaqueCloudCoveragePercents"] is None:
            del d["OpaqueCloudCoveragePercents"]
        if d["Temperatures"] is None:
            del d["Temperatures"]
        return d

    def is_valid(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid, errors = self.passes_derived_validations()
        return is_valid, errors

    def passes_derived_validations(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid = True
        errors = []
        if self.TypeName != "csv.weather.forecast.sync":
            is_valid = False
            errors.append(
                f"Payload requires TypeName of 'csv.weather.forecast.sync', not {self.TypeName}."
            )
        if not isinstance(self.LocationAlias, str):
            is_valid = False
            errors.append(f"LocationAlias must have type str.")
        if not isinstance(self.StartYearUtc, int):
            is_valid = False
            errors.append(f"StartYearUtc must have type int.")
        if not isinstance(self.StartMonthUtc, int):
            is_valid = False
            errors.append(f"StartMonthUtc must have type int.")
        if not isinstance(self.StartDayUtc, int):
            is_valid = False
            errors.append(f"StartDayUtc must have type int.")
        if not isinstance(self.StartHourUtc, int):
            is_valid = False
            errors.append(f"StartHourUtc must have type int.")
        try:
            datetime.datetime(
                year=self.StartYearUtc,
                month=self.StartMonthUtc,
                day=self.StartDayUtc,
                hour=self.StartHourUtc,
            )
        except ValueError as e:
            is_valid = False
            errors.append(e)
        if not isinstance(self.UniformSliceDurationHrs, float):
            is_valid = False
            errors.append(f"UniformSliceDurationHrs must have type float.")
        if not isinstance(self.TimezoneString, str):
            is_valid = False
            errors.append(f"TimezoneString {self.TimezoneString} must have type str.")
        if not isinstance(self.WeatherUid, str):
            is_valid = False
            errors.append(f"WeatherUid {self.WeatherUid} must have type str.")
        if not property_format.is_uuid_canonical_textual(self.WeatherUid):
            is_valid = False
            errors.append(
                f"WeatherUid {self.WeatherUid} must have format UuidCanonicalTextual."
            )

        return is_valid, errors


class CsvWeatherForecastSync_Maker:
    mp_alias = "csv.weather.forecast.sync"

    @classmethod
    def payload_is_valid(
        cls, payload_as_dict: Dict[str, Any]
    ) -> Tuple[bool, Optional[List[str]]]:
        try:
            payload_as_dict["UniformSliceDurationHrs"] = float(
                payload_as_dict["UniformSliceDurationHrs"]
            )
        except ValueError:
            pass  # This will get caught in is_valid() check below
        if "TempUnit" not in payload_as_dict.keys():
            payload_as_dict["TempUnit"] = None
        if "TempSourceAlias" not in payload_as_dict.keys():
            payload_as_dict["TempSourceAlias"] = None
        if "TempMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["TempMethodAlias"] = None
        if "IrradianceType" not in payload_as_dict.keys():
            payload_as_dict["IrradianceType"] = None
        if "IrradianceSourceAlias" not in payload_as_dict.keys():
            payload_as_dict["IrradianceSourceAlias"] = None
        if "IrradianceMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["IrradianceMethodAlias"] = None
        if "SkyClaritySourceAlias" not in payload_as_dict.keys():
            payload_as_dict["SkyClaritySourceAlias"] = None
        if "SkyClarityMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["SkyClarityMethodAlias"] = None
        if "Temperatures" not in payload_as_dict.keys():
            payload_as_dict["Temperatures"] = None
        if "IrradianceWPerM2" not in payload_as_dict.keys():
            payload_as_dict["IrradianceWPerM2"] = None
        if "OpaqueCloudCoveragePercents" not in payload_as_dict.keys():
            payload_as_dict["OpaqueCloudCoveragePercents"] = None

        try:
            p = CsvWeatherForecast(
                TypeName=payload_as_dict["TypeName"],
                LocationAlias=payload_as_dict["LocationAlias"],
                Comment=payload_as_dict["Comment"],
                StartYearUtc=payload_as_dict["StartYearUtc"],
                StartMonthUtc=payload_as_dict["StartMonthUtc"],
                StartDayUtc=payload_as_dict["StartDayUtc"],
                StartHourUtc=payload_as_dict["StartHourUtc"],
                UniformSliceDurationHrs=payload_as_dict["UniformSliceDurationHrs"],
                TimezoneString=payload_as_dict["TimezoneString"],
                ForecastUnixTimeS=payload_as_dict["ForecastUnixTimeS"],
                WeatherUid=payload_as_dict["WeatherUid"],
                TempUnit=payload_as_dict["TempUnit"],
                IrradianceType=payload_as_dict["IrradianceType"],
                TempSourceAlias=payload_as_dict["TempSourceAlias"],
                TempMethodAlias=payload_as_dict["TempMethodAlias"],
                IrradianceSourceAlias=payload_as_dict["IrradianceSourceAlias"],
                IrradianceMethodAlias=payload_as_dict["IrradianceMethodAlias"],
                SkyClaritySourceAlias=payload_as_dict["SkyClaritySourceAlias"],
                SkyClarityMethodAlias=payload_as_dict["SkyClarityMethodAlias"],
                Temperatures=payload_as_dict["Temperatures"],
                IrradianceWPerM2=payload_as_dict["IrradianceWPerM2"],
                OpaqueCloudCoveragePercents=payload_as_dict[
                    "OpaqueCloudCoveragePercents"
                ],
            )
        except TypeError:
            errors = [TypeError]
            return False, errors
        return p.is_valid()

    def __init__(
        self,
        weather_csv,
    ):
        self.errors = []
        self.tuple: Optional[CsvWeatherForecast] = None
        first_data_row = 22
        temperatures = []
        irradiances = []
        sky_clarities = []
        uniform_slice_duration_hrs = np.array([])
        with open(weather_csv, newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            line_idx = 0
            for row in reader:
                if line_idx == 0:
                    property = row[0].replace("\ufeff", "")
                    mp_alias = row[1]
                    if property != "TypeName":
                        raise Exception(
                            f"Csv 0,0 must be TypeName, instead it is {property}"
                        )
                    if mp_alias != "csv.weather.forecast.sync":
                        raise Exception(
                            f"TypeName for {weather_csv} must be csv.weather.forecast.sync. Instead it was {mp_alias}"
                        )
                if line_idx == 1:
                    property = row[0]
                    location_alias = row[1].strip()
                    if property != "LocationAlias":
                        raise Exception(
                            f"Csv 1,0 must be LocationAlias, instead it is {property}"
                        )
                if line_idx == 2:
                    property = row[0]
                    if len(row) == 1:
                        temp_source_alias = None
                    else:
                        temp_source_alias = row[1].strip()
                        if temp_source_alias == "":
                            temp_source_alias = None
                    if property != "TempSourceAlias":
                        raise Exception(
                            f"Csv 2,0 must be TempSourceAlias, instead it is {property}"
                        )
                if line_idx == 3:
                    property = row[0]
                    temp_method_alias = row[1].strip()
                    if len(row) == 1:
                        temp_method_alias = None
                    else:
                        if temp_method_alias == "":
                            temp_method_alias = None
                    if property != "TempMethodAlias":
                        raise Exception(
                            f"Csv 3,0 must be TempMethodAlias, instead it is {property}"
                        )
                if line_idx == 4:
                    property = row[0]
                    if len(row) == 1:
                        sky_clarity_source_alias = None
                    else:
                        sky_clarity_source_alias = row[1].strip()
                        if sky_clarity_source_alias == "":
                            sky_clarity_source_alias = None
                    if property != "SkyClaritySourceAlias":
                        raise Exception(
                            f"Csv 4,0 must be SkyClaritySourceAlias, instead it is {property}"
                        )
                if line_idx == 5:
                    property = row[0]
                    if len(row) == 1:
                        sky_clarity_method_alias = None
                    else:
                        sky_clarity_method_alias = row[1].strip()
                        if sky_clarity_method_alias == "":
                            sky_clarity_method_alias = None
                    if property != "SkyClarityMethodAlias":
                        raise Exception(
                            f"Csv 5,0 must be SkyClarityMethodAlias, instead it is {property}"
                        )
                if line_idx == 6:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_source_alias = None
                    else:
                        irradiance_source_alias = row[1].strip()
                        if irradiance_source_alias == "":
                            irradiance_source_alias = None
                    if property != "IrradianceSourceAlias":
                        raise Exception(
                            f"Csv 6,0 of {weather_csv} must be IrradianceSourceAlias, instead it is {property}"
                        )
                if line_idx == 7:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_method_alias = None
                    else:
                        irradiance_method_alias = row[1].strip()
                        if irradiance_method_alias == "":
                            irradiance_method_alias = None
                    if property != "IrradianceMethodAlias":
                        raise Exception(
                            f"Csv 7,0 must be IrradianceMethodAlias, instead it is {property}"
                        )
                if line_idx == 8:
                    property = row[0]
                    comment = row[1].strip()
                    if property != "Comment":
                        raise Exception(
                            f"Csv 8,0 must be Comment, instead it is {property}"
                        )
                if line_idx == 9:
                    property = row[0]
                    start_year_utc = int(row[1])
                    if property != "StartYearUtc":
                        raise Exception(
                            f"Csv 9,0 must be StartYearUtc, instead it is {property}"
                        )
                if line_idx == 10:
                    property = row[0]
                    start_month_utc = int(row[1])
                    if property != "StartMonthUtc":
                        raise Exception(
                            f"Csv 10,0 must be StartMonthUtc, instead it is {property}"
                        )
                if line_idx == 11:
                    property = row[0]
                    start_day_utc = int(row[1])
                    if property != "StartDayUtc":
                        raise Exception(
                            f"Csv 11,0 must be StartDayUtc, instead it is {property}"
                        )
                if line_idx == 12:
                    property = row[0]
                    start_hour_utc = int(row[1])
                    if property != "StartHourUtc":
                        raise Exception(
                            f"Csv 12,0 must be StartHourUtc, instead it is {property}"
                        )
                if line_idx == 13:
                    property = row[0]
                    uniform_slice_duration_hrs = float(row[1])
                    if property != "UniformSliceDurationHrs":
                        raise Exception(
                            f"Csv 13,0 must be UniformSliceDurationHrs, instead it is {property}"
                        )
                if line_idx == 14:
                    property = row[0]
                    timezone_string = row[1].strip()
                    if property != "TimezoneString":
                        raise Exception(
                            f"Csv 14,0 must be TimezoneString, instead it is {property}"
                        )
                if line_idx == 15:
                    property = row[0]
                    if len(row) == 1:
                        temp_unit = None
                    else:
                        temp_unit = row[1].strip()
                        if temp_unit == "":
                            temp_unit = None
                    if property != "TempUnit":
                        raise Exception(
                            f"Csv 15,0 must be TempUnit, instead it is {property}"
                        )
                if line_idx == 16:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_type = None
                    else:
                        irradiance_type = row[1].strip()
                        if irradiance_type == "":
                            irradiance_type = None
                    if property != "IrradianceType":
                        raise Exception(
                            f"Csv 16,0 must be IrradianceType, instead it is {property}"
                        )
                if line_idx == 17:
                    property = row[0]
                    include_temp_str = row[1].strip()
                    if property != "IncludeTemp":
                        raise Exception(
                            f"Csv 17,0 must be IncludeTemp, instead it is {property}"
                        )
                    if include_temp_str == "True":
                        include_temp = True
                    else:
                        include_temp = False
                if line_idx == 18:
                    property = row[0]
                    include_irradiance_str = row[1].strip()
                    if property != "IncludeIrradiance":
                        raise Exception(
                            f"Csv 18,0 must be IncludeIrradiance, instead it is {property}"
                        )
                    if include_irradiance_str == "True":
                        include_irradiance = True
                    else:
                        include_irradiance = False
                if line_idx == 19:
                    property = row[0]
                    include_sky_clarity_str = row[1].strip()
                    if property != "IncludeSkyClarity":
                        raise Exception(
                            f"Csv 19,0 must be IncludeSkyClarity, instead it is {property}"
                        )
                    if include_sky_clarity_str == "True":
                        include_sky_clarity = True
                    else:
                        include_sky_clarity = False
                if line_idx == 20:
                    property = row[0]
                    forecast_unix_time_s = int(row[1])
                    if property != "ForecastUnixTimeS":
                        raise Exception(
                            f"Csv 20,0 must be ForecastUnixTimeS, instead it is {property}"
                        )
                if line_idx == 21:
                    property = row[0]
                    price_uid = row[1].strip()
                    if property != "WeatherUid":
                        raise Exception(
                            f"Csv 21,0 must be WeatherUid, instead it is {property}"
                        )
                if line_idx == 22:
                    header0 = row[0].strip()
                    header1 = row[1].strip()
                    header2 = row[2].strip()
                    if header0 != "Temperatures":
                        raise Exception(
                            f"Csv 22,0 must be Temperatures, instead it is {header0}"
                        )
                    if header1 != "IrradianceWPerM2":
                        raise Exception(
                            f"Csv 22,1 must be IrradianceWPerM2, instead it is {header1}"
                        )
                    if header2 != "OpaqueCloudCoveragePercents":
                        raise Exception(
                            f"Csv 22,2 must be OpaqueCloudCoveragePercents, instead it is {header2}"
                        )
                elif line_idx >= first_data_row:
                    if include_temp:
                        try:
                            temperatures.append(int(row[0]))
                        except ValueError:
                            raise Exception(
                                f"File {weather_csv} Missing a temperature in row {line_idx+1}"
                            )
                    if include_irradiance:
                        try:
                            irradiances.append(int(row[1]))
                        except ValueError:
                            raise Exception(
                                f"Missing an irradiance in row {line_idx+1}"
                            )
                    if include_sky_clarity:
                        try:
                            sky_clarities.append(int(row[2]))
                        except ValueError:
                            raise Exception(f"Missing an occp in row {line_idx+1}")
                line_idx += 1
        if not include_temp:
            temperatures = None
            temp_unit = None
        if not include_irradiance:
            irradiances = None
            irradiance_type = None
        if not include_sky_clarity:
            sky_clarities = None

        p = CsvWeatherForecast(
            LocationAlias=location_alias,
            Comment=comment,
            StartYearUtc=start_year_utc,
            StartMonthUtc=start_month_utc,
            StartDayUtc=start_day_utc,
            StartHourUtc=start_hour_utc,
            UniformSliceDurationHrs=uniform_slice_duration_hrs,
            TimezoneString=timezone_string,
            ForecastUnixTimeS=forecast_unix_time_s,
            WeatherUid=price_uid,
            TempUnit=temp_unit,
            IrradianceType=irradiance_type,
            TempSourceAlias=temp_source_alias,
            TempMethodAlias=temp_method_alias,
            IrradianceSourceAlias=irradiance_source_alias,
            IrradianceMethodAlias=irradiance_method_alias,
            SkyClaritySourceAlias=sky_clarity_source_alias,
            SkyClarityMethodAlias=sky_clarity_method_alias,
            Temperatures=temperatures,
            IrradianceWPerM2=irradiances,
            OpaqueCloudCoveragePercents=sky_clarities,
        )

        is_valid, errors = p.is_valid()
        if is_valid is False:
            self.errors = errors
            print(errors)
            raise SchemaError(
                f"Failed to create payload due to these errors:{errors}. Input file is {weather_csv}"
            )

        self.tuple = p



import csv
import json
from typing import List

from gridworks.algo_utils import BasicAccount

from gwatn.types import TadeedSpecsHack_Maker


defaults_file = "input_data/ta_nft_data.csv"
rows: List[str] = []
with open(defaults_file) as csv_file:
    reader = csv.reader(csv_file)
    for row in reader:
        rows.append(row)


plant_names: List[str] = []
for i in range(1, len(rows)):
    plant = rows[i][0].strip()
    plant_names.append(plant)

dups = []
pnames = []
for p in plant_names:
    if p in pnames:
        dups.append(p)
    else:
        pnames.append(p)

if len(dups) > 0:
    raise Exception(f"Duplicate plant names: {dups}")

for i in range(1, len(rows)):
    plant = rows[i][0].strip()
    lat = int(rows[i][1])
    lon = int(rows[i][2])
    port: str = rows[i][3].strip()
    ta_alias = f"d1.isone.ver.keene.{plant}.ta"
    payload = TadeedSpecsHack_Maker(
        terminal_asset_alias=ta_alias, micro_lat=lat, micro_lon=lon, daemon_port=port
    ).tuple
    file_name = (
        f"input_data/eventstore/d1.isone.ver.keene.{plant}.ta-tadeed.specs.hack.json"
    )
    with open(file_name, "w") as f:
        f.write(json.dumps(payload.as_dict()))
